# -*- coding: utf-8 -*-
import six
from odin import bases
from odin import exceptions
from odin.resources import create_resource_from_dict
from odin.fields import Field
from odin.utils import value_in_choices, getmeta
from odin.validators import EMPTY_VALUES

__all__ = ('CompositeField', 'DictAs', 'ObjectAs', 'ListOf', 'ArrayOf', 'DictOf')


class CompositeField(Field):
    """
    The base class for composite (or fields that contain other resources) eg DictAs/ListOf fields.
    """
    def __init__(self, resource, use_container=False, **options):
        """
        Initialisation of a CompositeField.

        :param resource:
        :param use_container: Special flag for codecs that support containers or just multiple instances of a
            sub element (ie XML).
        :param empty: This collection can be empty
        :param options: Additional options passed to :py:class:`odin.fields.Field` super class.

        """
        if not hasattr(resource, '_meta'):
            raise TypeError("``%r`` is not a valid type for a related field." % resource)
        self.of = resource
        self.use_container = use_container

        if not options.get('null', False):
            options.setdefault('default', lambda: resource())

        super(CompositeField, self).__init__(**options)

    def to_python(self, value):
        if value is None:
            return None
        if isinstance(value, self.of):
            return value
        if isinstance(value, dict):
            return create_resource_from_dict(value, getmeta(self.of).resource_name)
        msg = self.error_messages['invalid'] % self.of
        raise exceptions.ValidationError(msg)

    def validate(self, value):
        super(CompositeField, self).validate(value)
        if value not in EMPTY_VALUES:
            value.full_clean()

    def item_iter_from_object(self, obj):
        """
        Return an iterator of items (resource, idx) from composite field.

        For single items (eg ``DictAs`` will return a list a single item (resource, None))

        :param obj:
        :return:
        """
        raise NotImplementedError()

    def key_to_python(self, key):
        """
        A to python method for the key value.
        :param key:
        :return:
        """
        raise NotImplementedError()


class DictAs(CompositeField):
    """
    Treat a dictionary as a Resource.
    """
    default_error_messages = {
        'invalid': "Must be a dict of type ``%r``.",
    }
    data_type_name = "Dict as"

    def item_iter_from_object(self, obj):
        resource = self.value_from_object(obj)
        if resource:
            yield (None, resource)

    def key_to_python(self, key):
        pass  # Not required as keys are not used.

ObjectAs = DictAs


class ListOf(CompositeField):
    """
    List of resources.
    """
    default_error_messages = {
        'invalid': "Must be a list of ``%r`` objects.",
        'null': "List cannot contain null entries.",
        'empty': "List cannot be empty",
    }
    data_type_name = "List of"

    def __init__(self, resource, empty=True, **options):
        options.setdefault('default', list)
        super(ListOf, self).__init__(resource, **options)
        self.empty = empty

    @staticmethod
    def _process_list(value_list, method):
        values = []
        errors = {}
        for idx, value in enumerate(value_list):
            error_key = str(idx)

            try:
                values.append(method(value))
            except exceptions.ValidationError as ve:
                errors[error_key] = ve.error_messages

        if errors:
            raise exceptions.ValidationError(errors)

        return values

    def to_python(self, value):
        if value is None:
            return None
        if isinstance(value, (list, bases.ResourceIterable)):
            super_to_python = super(ListOf, self).to_python

            def process(val):
                if val is None:
                    raise exceptions.ValidationError(self.error_messages['null'])
                return super_to_python(val)

            return self._process_list(value, process)
        msg = self.error_messages['invalid'] % self.of
        raise exceptions.ValidationError(msg)

    def validate(self, value):
        # Skip The direct super method and apply it to each list item.
        super(CompositeField, self).validate(value)  # noqa
        if value is not None:
            super_validate = super(ListOf, self).validate
            self._process_list(value, super_validate)

        if (value is not None) and (not value) and (not self.empty):
            raise exceptions.ValidationError(self.error_messages['empty'])

    def __iter__(self):
        # This does nothing but it does prevent inspections from complaining.
        return None  # NoQA

    def item_iter_from_object(self, obj):
        resources = self.value_from_object(obj)
        if resources:
            for i in enumerate(resources):
                yield i

    def key_to_python(self, key):
        """
        A to python method for the key value.
        :param key:
        :return:
        """
        return int(key)

ArrayOf = ListOf


class DictOf(CompositeField):
    """
    Dictionary of resources.
    """
    default_error_messages = {
        'invalid': "Must be a dict of ``%r`` objects.",
        'null': "Dict cannot contain null entries.",
        'empty': "List cannot be empty",
        'invalid_key': 'Key %r is not a valid choice.',
    }
    data_type_name = "Dict of"

    def __init__(self, resource, empty=True, key_choices=None, **options):
        options.setdefault('default', dict)
        super(DictOf, self).__init__(resource, **options)
        self.empty = empty
        self.key_choices = key_choices

    def _process_dict(self, value_dict, method):
        values = {}
        errors = {}
        key_choices = self.key_choices
        for key, value in six.iteritems(value_dict):
            if key_choices and not value_in_choices(key, key_choices):
                msg = self.error_messages['invalid_key'] % value
                raise exceptions.ValidationError(msg)

            try:
                values[key] = method(value)
            except exceptions.ValidationError as ve:
                errors[key] = ve.error_messages

        if errors:
            raise exceptions.ValidationError(errors)

        return values

    def to_python(self, value):
        if value is None:
            return None
        if isinstance(value, dict):
            super_to_python = super(DictOf, self).to_python

            def process(val):
                if val is None:
                    raise exceptions.ValidationError(self.error_messages['null'], code='null')
                return super_to_python(val)

            return self._process_dict(value, process)
        msg = self.error_messages['invalid'] % self.of
        raise exceptions.ValidationError(msg)

    def validate(self, value):
        # Skip The direct super method and apply it to each list item.
        super(CompositeField, self).validate(value)  # noqa
        if value is not None:
            super_validate = super(DictOf, self).validate
            self._process_dict(value, super_validate)

        if (value is not None) and (not value) and (not self.empty):
            raise exceptions.ValidationError(self.error_messages['empty'], code='empty')

    def __iter__(self):
        # This does nothing but it does prevent inspections from complaining.
        return None  # NoQA

    def item_iter_from_object(self, obj):
        resources = self.value_from_object(obj)
        if resources:
            for key in sorted(resources):
                yield key, resources[key]

    def key_to_python(self, key):
        """
        A to python method for the key value.
        :param key:
        :return:
        """
        return key



# -*- coding: utf-8 -*-


def sum_fields(*field_values):
    """
    Return a sum of fields

    :param field_values:
    :return:

    """
    return sum(field_values)


class CatFields(object):
    """
    Helper for combining multiple fields
    """
    __slots__ = ('sep',)

    def __init__(self, sep=''):
        self.sep = sep

    def __call__(self, *field_values):
        return self.sep.join(field_values)

cat_fields = CatFields


class SplitField(object):
    """
    Helper for splitting a field into multiple fields.
    """
    __slots__ = ('sep', 'max_split',)

    def __init__(self, sep=None, max_split=None):
        self.sep = sep
        self.max_split = max_split

    def __call__(self, field_value):
        if self.max_split is None:
            return field_value.split(self.sep)
        else:
            return field_value.split(self.sep, self.max_split)

split_field = SplitField


class ApplyMapping(object):
    """
    Helper for applying a mapper.

    This helper should be used along with the bind flag so the context object can be maintained.
    """
    def __init__(self, mapping, allow_subclass=False):
        self.mapping = mapping
        self.allow_subclass = allow_subclass

    def __call__(self, bound_self, field_value):
        # Note this returns a tuple, this is expected
        return self.mapping.apply(field_value, context=bound_self.context, allow_subclass=self.allow_subclass),

    def __repr__(self):
        return "<%s: %s.%s>" % (self.__class__.__name__, self.mapping.__module__, self.mapping.__name__)

MapDictAs = MapListOf = ApplyMapping


class NoOpMapper(object):
    """
    Helper that provides the mapper interface performs no operation on the object.

    This is used with the MapListOf and MapDictAs fields when both contain the same Resource type.
    """
    @staticmethod
    def apply(source_obj, **_):
        return source_obj



# -*- coding: utf-8 -*-
# This file is largely verbatim from the Django project, the wheel works well, no need to re-invent it.
#
# A note: to use validators from the Django project install the baldr package. Baldr is an integration between Odin and
# the Django framework, the integration includes support for handling the Django version of the ValidationError
# exception within Odin.
import re
import six

from odin import exceptions
from odin.utils.ipv6 import is_valid_ipv6_address

EMPTY_VALUES = (None, '', [], (), {})


class RegexValidator(object):
    regex = r''
    message = 'Enter a valid value.'
    code = 'invalid'

    def __init__(self, regex=None, message=None, code=None):
        if regex is not None:
            self.regex = regex
        if message is not None:
            self.message = message
        if code is not None:
            self.code = code

        # Compile the regex if it was not passed pre-compiled.
        if isinstance(self.regex, six.string_types):
            self.regex = re.compile(self.regex)

    def __call__(self, value):
        """
        Validates that the input matches the regular expression.
        """
        if not self.regex.search(value):
            raise exceptions.ValidationError(self.message, code=self.code)


class URLValidator(RegexValidator):
    regex = re.compile(
        r'^(?:http|ftp)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|'  # ...or ipv4
        r'\[?[A-F0-9]*:[A-F0-9:]+\]?)'  # ...or ipv6
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)
    message = "Enter a valid URL value."

validate_url = URLValidator()


class BaseValidator(object):
    message = 'Ensure this value is %(limit_value)s (it is %(show_value)s).'
    code = 'limit_value'
    description = 'Ensure that a value is %(limit_value)s.'

    def __init__(self, limit_value):
        self.limit_value = limit_value

    def __call__(self, value):
        cleaned = self.clean(value)
        params = {'limit_value': self.limit_value, 'show_value': cleaned}
        if self.compare(cleaned, self.limit_value):
            raise exceptions.ValidationError(self.message % params, code=self.code, params=params)

    def __str__(self):
        return self.description % {'limit_value': self.limit_value}

    def compare(self, a, b):
        return a is not b

    def clean(self, value):
        return value


class MaxValueValidator(BaseValidator):
    message = 'Ensure this value is less than or equal to %(limit_value)s.'
    code = 'max_value'
    description = 'Ensure value is less than or equal to %(limit_value)s.'

    def compare(self, a, b):
        return a > b


class MinValueValidator(BaseValidator):
    message = 'Ensure this value is greater than or equal to %(limit_value)s.'
    code = 'min_value'
    description = 'Ensure value is greater than or equal to %(limit_value)s.'

    def compare(self, a, b):
        return a < b


class LengthValidator(BaseValidator):
    message = 'Ensure this value has exactly %(limit_value)d characters (it has %(show_value)d).'
    code = 'length'
    description = 'Ensure value has exactly %(limit_value)d characters.'

    def compare(self, a, b):
        return a != b

    def clean(self, value):
        return len(value)


class MaxLengthValidator(LengthValidator):
    message = 'Ensure this value has at most %(limit_value)d characters (it has %(show_value)d).'
    code = 'max_length'
    description = 'Ensure value has at most %(limit_value)d characters.'

    def compare(self, a, b):
        return a > b


class MinLengthValidator(LengthValidator):
    message = 'Ensure this value has at least %(limit_value)d characters (it has %(show_value)d).'
    code = 'min_length'
    description = 'Ensure value has at least %(limit_value)d characters.'

    def compare(self, a, b):
        return a < b


def simple_validator(assertion=None, message='The supplied value is invalid', code='invalid'):
    """
    Create a simple validator.

    :param assertion: An Validation exception will be raised if this check returns a none True value.
    :param message: Message to raised in Validation exception if validation fails.
    :param code: Code to included in Validation exception. This can be used to customise the message at the resource
        level.

    Usage::

        >>> none_validator = simple_validator(lambda x: x is not None, message="This value cannot be none")

    This can also be used as a decorator::

        @simple_validator(message="This value cannot be none")
        def none_validator(v):
            return v is not None
    """
    def inner(func):
        def wrapper(value):
            params = {'show_value': value}
            if not func(value):
                raise exceptions.ValidationError(message % params, code=code, params=params)
        return wrapper

    if assertion:
        return inner(assertion)
    else:
        return inner


class IPv4Address(RegexValidator):
    """
    Validate an IPv4 address
    """
    regex = r'^(25[0-5]|2[0-4][0-9]|[0-1]?[0-9]?[0-9])(\.(25[0-5]|2[0-4][0-9]|[0-1]?[0-9]?[0-9])){3}\Z'
    message = 'Enter a valid IPv4 address.'

validate_ipv4_address = IPv4Address()


@simple_validator(message="Enter a valid IPv6 address")
def validate_ipv6_address(value):
    """
    Validate an IPv6 address
    """
    return is_valid_ipv6_address(value)


def validate_ipv46_address(value):
    """
    Validate is either an IPv4 or IPv6 address
    """
    try:
        validate_ipv4_address(value)
    except exceptions.ValidationError:
        try:
            validate_ipv6_address(value)
        except exceptions.ValidationError:
            raise exceptions.ValidationError('Enter a valid IPv4 or IPv6 address.', code='invalid')


class EmailValidator(object):
    """
    Validate an Email address.
    """
    message = 'Enter a valid email address.'
    code = 'invalid'
    user_regex = re.compile(
        r"(^[-!#$%&'*+/=?^_`{}|~0-9A-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*\Z"  # dot-atom
        r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-\011\013\014\016-\177])*"\Z)',  # quoted-string
        re.IGNORECASE)
    domain_regex = re.compile(
        # max length for domain name labels is 63 characters per RFC 1034
        r'((?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+)(?:[A-Z0-9-]{2,63}(?<!-))\Z',
        re.IGNORECASE)
    literal_regex = re.compile(
        # literal form, ipv4 or ipv6 address (SMTP 4.1.3)
        r'\[([A-f0-9:.]+)\]\Z',
        re.IGNORECASE)
    domain_whitelist = ['localhost']

    def __init__(self, message=None, code=None, whitelist=None):
        if message is not None:
            self.message = message
        if code is not None:
            self.code = code
        if whitelist is not None:
            self.domain_whitelist = whitelist

    def __call__(self, value):
        if not value or '@' not in value:
            raise exceptions.ValidationError(self.message, code=self.code)

        user_part, domain_part = value.rsplit('@', 1)

        if not self.user_regex.match(user_part):
            raise exceptions.ValidationError(self.message, code=self.code)

        if (domain_part not in self.domain_whitelist and
                not self.validate_domain_part(domain_part)):
            # Try for possible IDN domain-part
            try:
                domain_part = domain_part.encode('idna').decode('ascii')
                if self.validate_domain_part(domain_part):
                    return
            except UnicodeError:
                pass
            raise exceptions.ValidationError(self.message, code=self.code)

    def validate_domain_part(self, domain_part):
        if self.domain_regex.match(domain_part):
            return True

        literal_match = self.literal_regex.match(domain_part)
        if literal_match:
            ip_address = literal_match.group(1)
            try:
                validate_ipv46_address(ip_address)
                return True
            except exceptions.ValidationError:
                pass

validate_email_address = EmailValidator()




class BaseField(object):
    # These track each time an instance is created. Used to retain order.
    creation_counter = 0

    def __init__(self, verbose_name=None, verbose_name_plural=None, name=None, doc_text=''):
        self.verbose_name, self.verbose_name_plural = verbose_name, verbose_name_plural
        self.name = name
        self.doc_text = doc_text

        self.creation_counter = BaseField.creation_counter
        BaseField.creation_counter += 1

        self.attname = None

    def __hash__(self):
        return self.creation_counter

    def __repr__(self):
        """
        Displays the module, class and name of the field.
        """
        path = '%s.%s' % (self.__class__.__module__, self.__class__.__name__)
        name = getattr(self, 'name', None)
        if name is not None:
            return '<%s: %s>' % (path, name)
        return '<%s>' % path

    def set_attributes_from_name(self, attname):
        if not self.name:
            self.name = attname
        self.attname = attname
        if self.verbose_name is None and self.name:
            self.verbose_name = self.name.replace('_', ' ')
        if self.verbose_name_plural is None and self.verbose_name:
            self.verbose_name_plural = "%ss" % self.verbose_name

    def prepare(self, value):
        """
        Prepare a value for serialisation.

        :param value:
        :return:
        """
        return value

    def as_string(self, value):
        """
        Generate a string representation of a field.

        :param value:
        :rtype: str

        """
        if value is not None:
            return str(value)

    def value_from_object(self, obj):
        """
        Returns the value of this field in the given resource instance.
        """
        return getattr(obj, self.attname)



# -*- coding: utf-8 -*-
import pytest
from odin.contrib.geo.datatypes import latitude, longitude, latlng, point
from odin.contrib.geo.fields import LatitudeField, LongitudeField, LatLngField, PointField
from odin.exceptions import ValidationError


class TestGeoFields(object):
    # LatitudeField ###########################################################

    def test_latitude_field(self):
        f = LatitudeField()
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, 'a')
        pytest.raises(ValidationError, f.clean, 92)
        pytest.raises(ValidationError, f.clean, -92)
        assert 12.345 == f.clean(12.345)
        assert -12.345 == f.clean(-12.345)
        assert 12.345 == f.clean('12.345')
        assert isinstance(f.clean(12.345), latitude)

    # LongitudeField ##########################################################

    def test_longitude_field(self):
        f = LongitudeField()
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, 'a')
        pytest.raises(ValidationError, f.clean, 182)
        pytest.raises(ValidationError, f.clean, -182)
        assert 123.456 == f.clean(123.456)
        assert -123.456 == f.clean(-123.456)
        assert 123.456 == f.clean('123.456')
        assert isinstance(f.clean(122.456), longitude)

    # LatLngField #############################################################

    def test_latlng_field(self):
        f = LatLngField()
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, 'a')
        pytest.raises(ValidationError, f.clean, [])
        pytest.raises(ValidationError, f.clean, [12.345])
        pytest.raises(ValidationError, f.clean, [92, 123.345])
        pytest.raises(ValidationError, f.clean, [12.345, 182])

        assert (12.345, 123.456) == f.clean((12.345, 123.456))
        assert (12.345, 123.456) == f.clean(('12.345', '123.456'))
        assert isinstance(f.clean((12.345, 123.456)), latlng)

    # PointField ##############################################################

    def test_point_field(self):
        f = PointField()
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, 'a')
        pytest.raises(ValidationError, f.clean, [])
        pytest.raises(ValidationError, f.clean, [1])
        pytest.raises(ValidationError, f.clean, [1, 2, 3, 4])
        pytest.raises(ValidationError, f.clean, [1, 2, 'a', 4])

        assert (1, 2) == f.clean((1, 2))
        assert (1, 2) == f.clean(('1', '2'))
        assert (1, 2, 3) == f.clean((1, 2, 3))
        assert (1, 2, 3) == f.clean(('1', '2', '3'))
        assert isinstance(f.clean((1, 2)), point)
        assert isinstance(f.clean((1, 2, 3)), point)



# -*- coding: utf-8 -*-
import six
import datetime

from odin.utils import getmeta

from odin import bases
from odin import serializers, resources, ResourceAdapter
from odin.exceptions import CodecDecodeError, CodecEncodeError

try:
    import simplejson as json
except ImportError:
    import json

LIST_TYPES = [bases.ResourceIterable]
if six.PY3:
    import typing
    LIST_TYPES += [typing.ValuesView, typing.KeysView]
LIST_TYPES = tuple(LIST_TYPES)

JSON_TYPES = {
    datetime.date: serializers.date_iso_format,
    datetime.time: serializers.time_iso_format,
    datetime.datetime: serializers.datetime_iso_format,
}
CONTENT_TYPE = 'application/json'


class OdinEncoder(json.JSONEncoder):
    """
    Encoder for Odin resources.
    """
    def __init__(self, include_virtual_fields=True, include_type_field=True, *args, **kwargs):
        super(OdinEncoder, self).__init__(*args, **kwargs)
        self.include_virtual_fields = include_virtual_fields
        self.include_type_field = include_type_field

    def default(self, o):
        if isinstance(o, (resources.ResourceBase, ResourceAdapter)):
            meta = getmeta(o)
            obj = o.to_dict(self.include_virtual_fields)
            if self.include_type_field:
                obj[meta.type_field] = meta.resource_name
            return obj
        elif isinstance(o, LIST_TYPES):
            return list(o)
        elif o.__class__ in JSON_TYPES:
            return JSON_TYPES[o.__class__](o)
        return super(OdinEncoder, self).default(o)


def load(fp, resource=None, full_clean=True, default_to_not_supplied=False):
    """
    Load a from a JSON encoded file.

    See :py:meth:`loads` for more details of the loading operation.

    :param fp: a file pointer to read JSON data from.
    :param resource: A resource type, resource name or list of resources and names to use as the base for creating a
        resource. If a list is supplied the first item will be used if a resource type is not supplied.
    :param full_clean: Do a full clean of the object as part of the loading process.
    :param default_to_not_supplied: Used for loading partial resources. Any fields not supplied are replaced with
        NOT_SUPPLIED.
    :returns: A resource object or object graph of resources loaded from file.

    """
    return loads(fp.read(), resource, full_clean, default_to_not_supplied)


def loads(s, resource=None, full_clean=True, default_to_not_supplied=False):
    """
    Load from a JSON encoded string.

    If a ``resource`` value is supplied it is used as the base resource for the supplied JSON. I one is not supplied a
    resource type field ``$`` is used to obtain the type represented by the dictionary. A ``ValidationError`` will be
    raised if either of these values are supplied and not compatible. It is valid for a type to be supplied in the file
    to be a child object from within the inheritance tree.

    :param s: String to load and parse.
    :param resource: A resource type, resource name or list of resources and names to use as the base for creating a
        resource. If a list is supplied the first item will be used if a resource type is not supplied.
    :param full_clean: Do a full clean of the object as part of the loading process.
    :param default_to_not_supplied: Used for loading partial resources. Any fields not supplied are replaced with
        NOT_SUPPLIED.
    :returns: A resource object or object graph of resources parsed from supplied string.

    """
    try:
        return resources.build_object_graph(json.loads(s), resource, full_clean, False, default_to_not_supplied)
    except (ValueError, TypeError) as ex:
        raise CodecDecodeError(str(ex))


def dump(resource, fp, cls=OdinEncoder, **kwargs):
    """
    Dump to a JSON encoded file.

    :param resource: The root resource to dump to a JSON encoded file.
    :param cls: Encoder to use serializing to a string; default is the :py:class:`OdinEncoder`.
    :param fp: The file pointer that represents the output file.

    """
    try:
        json.dump(resource, fp, cls=cls, **kwargs)
    except ValueError as ex:
        raise CodecEncodeError(str(ex))


def dumps(resource, cls=OdinEncoder, **kwargs):
    """
    Dump to a JSON encoded string.

    :param resource: The root resource to dump to a JSON encoded file.
    :param cls: Encoder to use serializing to a string; default is the :py:class:`OdinEncoder`.
    :returns: JSON encoded string.

    """
    try:
        return json.dumps(resource, cls=cls, **kwargs)
    except ValueError as ex:
        raise CodecEncodeError(str(ex))



# -*- coding: utf-8 -*-
import pytest
from odin import validators
from odin.exceptions import ValidationError


class TestValidator(object):
    def test_regex_validator(self):
        target = validators.RegexValidator(r'^[a-z]{3}$', "Please enter 3 alpha characters.", "match_chars")

        assert "Please enter 3 alpha characters." == target.message
        assert "match_chars" == target.code

        target("abc")
        target("cba")
        pytest.raises(ValidationError, target, "a")
        pytest.raises(ValidationError, target, "abcd")
        pytest.raises(ValidationError, target, "123")

    def test_url(self):
        validators.validate_url('http://www.djangoproject.com/')
        validators.validate_url('http://localhost/')
        validators.validate_url('http://example.com/')
        validators.validate_url('http://www.example.com/')
        validators.validate_url('http://www.example.com:8000/test')
        validators.validate_url('http://valid-with-hyphens.com/')
        validators.validate_url('http://subdomain.example.com/')
        validators.validate_url('http://200.8.9.10/')
        validators.validate_url('http://200.8.9.10:8000/test')
        validators.validate_url('http://valid-----hyphens.com/')
        validators.validate_url('http://example.com?something=value')
        validators.validate_url('http://example.com/index.php?something=value&another=value2')
        validators.validate_url('https://example.com/')
        validators.validate_url('ftp://example.com/')
        validators.validate_url('ftps://example.com/')
        validators.validate_url('http://savage.company/')
    
        pytest.raises(ValidationError, validators.validate_url, 'foo')
        pytest.raises(ValidationError, validators.validate_url, 'http://')
        pytest.raises(ValidationError, validators.validate_url, 'http://example')
        pytest.raises(ValidationError, validators.validate_url, 'http://example.')
        pytest.raises(ValidationError, validators.validate_url, 'http://.com')
        pytest.raises(ValidationError, validators.validate_url, 'http://invalid-.com')
        pytest.raises(ValidationError, validators.validate_url, 'http://-invalid.com')
        pytest.raises(ValidationError, validators.validate_url, 'http://inv-.alid-.com')
        pytest.raises(ValidationError, validators.validate_url, 'http://inv-.-alid.com')

    def test_max_value_validator(self):
        target = validators.MaxValueValidator(10)

        target(10)
        target(1)
        target(-10)
        pytest.raises(ValidationError, target, 11)

    def test_min_value_validator(self):
        target = validators.MinValueValidator(10)

        target(10)
        pytest.raises(ValidationError, target, 1)
        pytest.raises(ValidationError, target, -10)
        target(11)

    def test_length_validator(self):
        target = validators.LengthValidator(10)
        target("1234567890")
        pytest.raises(ValidationError, target, "123456789")
        pytest.raises(ValidationError, target, "12345678901")

    def test_max_length_validator(self):
        target = validators.MaxLengthValidator(10)

        target("123457890")
        target("12345")
        target("")
        pytest.raises(ValidationError, target, "12345678901")

    def test_min_length_validator(self):
        target = validators.MinLengthValidator(10)

        pytest.raises(ValidationError, target, "123457890")
        pytest.raises(ValidationError, target, "12345")
        pytest.raises(ValidationError, target, "")
        target("12345678901")


class TestSimpleValidator(object):
    def test_method(self):
        def reflect(v):
            return v

        validator = validators.simple_validator(reflect)

        pytest.raises(ValidationError, validator, False)
        validator(True)

    def test_decorator(self):

        @validators.simple_validator()
        def reflect_validator(v):
            return v

        pytest.raises(ValidationError, reflect_validator, False)
        reflect_validator(True)


@pytest.mark.parametrize('value', (
    '192.168.0.1',
    '1.1.1.1',
    '255.255.255.255',
))
def test_ipv4_address_valid(value):
    validators.validate_ipv4_address(value)


@pytest.mark.parametrize('value', (
    '.2.3.4',
    '1.2.3',
    '1.2.3.4.5',
    '300.0.0.1',
    'ff.ff.ff.ff',
    'abcd',
    '',
    '::ff:ff:ff:ff'
))
def test_ipv4_address_invalid(value):
    with pytest.raises(ValidationError):
        validators.validate_ipv4_address(value)


@pytest.mark.parametrize('value', (
    '::',  # Localhost
    '::1',  # Default unicast
    '::ffff:0:0',  # IPv4 Mapped address
    '2002::',  # 6to4
    '2001::',  # Teredo tunneling
    'fc00::',  # Unique local address
    'ffff::192.168.0.1',
    '1:2:3:4:5:6:7:8',
    'dead:beef:dead:beef:dead:beef:dead:beef'
))
def test_ipv6_address_valid(value):
    validators.validate_ipv6_address(value)


@pytest.mark.parametrize('value', (
    '192.168.0.1',
    '1:2:3:4:5:6:7:8:9',
    '1234:5678:9abc:defg::',
    '2001::12345',
    '::ffff::',
    'ff:::0',
    ':2:3:4:5:6:7:8',
    '1:2:3:4:5:6:7:',
    ':2:3:4:5:6:7:',
    '1:2:3:4:5:192.168.0',
    '1:2:3:4:256.256.256.256',
    '192.168.0.1::ff',
    'zzzz::0',
))
def test_ipv6_address_invalid(value):
    with pytest.raises(ValidationError):
        validators.validate_ipv6_address(value)


@pytest.mark.parametrize('value', (
    '192.168.0.1',
    '1:2:3:4:5:6:7:8',
))
def test_validate_ipv46_address_valid(value):
    validators.validate_ipv46_address(value)


@pytest.mark.parametrize('value', (
    '192.168.0.',
    '1:2:3:4:5:6:7:8:9',
))
def test_validate_ipv46_address_invalid(value):
    with pytest.raises(ValidationError):
        validators.validate_ipv46_address(value)


@pytest.mark.parametrize('value kwargs'.split(), (
    ('foo@bar.com', None),
    ('FOO@BAR.COM', None),
    ('foo@localhost', None),
    # New release domain names
    ('foo@example.company', None),
    # IPv4 based email
    ('foo@127.0.0.1', None),
    # IPv6 based email
    ('foo@::1', None),
    ('foo@[::1]', None),
))
def test_validate_email_valid(value, kwargs):
    kwargs = kwargs or dict()
    validators.EmailValidator(**kwargs)(value)


@pytest.mark.parametrize('value kwargs'.split(), (
    ('foo', None),
    ('.foo@bar.com', None),
    ('foo@corpau', None),
    ('foo@localhost', {'whitelist': ''}),
    ('foo@{0}.com'.format('a' * 64), None),
    ('foo@[::ff::0]', None),
))
def test_validate_email_valid(value, kwargs):
    kwargs = kwargs or dict()
    with pytest.raises(ValidationError):
        validators.EmailValidator(**kwargs)(value)



from __future__ import absolute_import, print_function

from humanfriendly.terminal import ansi_wrap, HIGHLIGHT_COLOR

from odin.utils import getmeta


def resource_reference(resource):
    return "{} <{}.{}>".format(getmeta(resource).resource_name, resource.__module__, resource.__name__)


class SummaryBase(object):
    """
    Base class for summaries
    """
    def __init__(self, out):
        self.out = out

    def print(self, objects, sep=' ', end='\n'):
        print(*objects, sep=sep, end=end, file=self.out)

    def title(self, value, underline='='):
        self.print([ansi_wrap(value, bold=True, color=HIGHLIGHT_COLOR), underline * len(value)], sep='\n')



# -*- coding: utf-8 -*-
import odin
from odin.contrib.money import AmountField
from odin.contrib.money import Amount
from odin.codecs import json_codec


class AmountResource(odin.Resource):
    class Meta:
        namespace = "odin.tests"

    a = AmountField(null=True)
    b = AmountField()
    c = AmountField()


class TestMoneySerialisation(object):
    def test_serialise(self):
        resource = AmountResource(
            a=None,
            b=Amount(10),
            c=Amount(22.02, 'AUD')
        )

        actual = json_codec.dumps(resource, sort_keys=True)

        assert (
            actual ==
            '{"$": "odin.tests.AmountResource", "a": null, "b": [10.0, "XXX"], "c": [22.02, "AUD"]}'
        )

    def test_deserialise(self):
        resource = json_codec.loads(
            '{"$": "odin.tests.AmountResource", "a": null, "b": 10, "c": [23.66, "AUD"]}'
        )

        assert None == resource.a
        assert Amount(10) == resource.b
        assert Amount(23.66, "AUD") == resource.c



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import pytest
import datetime
from odin.exceptions import MappingSetupError, MappingExecutionError
from odin.mapping import MappingResult, FieldMapping
from odin.mapping.helpers import MapDictAs, MapListOf, NoOpMapper
from .resources import *


class SimpleFromResource(odin.Resource):
    title = odin.StringField()


class SimpleToResource(odin.Resource):
    title = odin.StringField()
    title_count = odin.StringField()


class FakeToResource(odin.Resource):
    title = odin.StringField()
    name = odin.StringField()


class SimpleFromTo(odin.Mapping):
    from_obj = SimpleFromResource
    to_obj = SimpleToResource

    @odin.map_field(from_field='title')
    def title_count(self, value):
        if self.in_loop:
            return "%s: %s" % (self.loop_idx, value)
        else:
            return value


def assertMappingEquivalent(a, b):
    flat_a = sorted(str(i) for i in a)
    flat_b = sorted(str(i) for i in b)
    assert flat_a == flat_b


class TestMappingBase(object):
    maxDiff = None

    def test_full_mapping(self):
        assertMappingEquivalent([
            FieldMapping(('from_field1',), None, ('to_field1',), False, False, False),
            FieldMapping(('from_field2',), int, ('to_field2',), False, False, False),
            FieldMapping(('from_field3', 'from_field4'), sum_fields, ('to_field3',), False, False, False),
            FieldMapping(('from_field1',), None, ('same_but_different',), False, False, False),
            FieldMapping(('from_field_c1', 'from_field_c2', 'from_field_c3'), 'multi_to_one', ('to_field_c1',), False, False, False),
            FieldMapping(('from_field_c4',), 'one_to_multi', ('to_field_c2', 'to_field_c3'), False, False, False),
            FieldMapping(('not_auto_c5',), 'not_auto_c5', ('not_auto_c5',), False, False, False),
            FieldMapping(('comma_separated_string',), 'comma_separated_string', ('array_string',), True, False, False),
            FieldMapping(None, 'assigned_field', ('assigned_field',), False, False, False),
            FieldMapping(('count',), 'default_action', ('count',), False, False, False),
            FieldMapping(('child',), MapDictAs(NoOpMapper), ('child',), False, True, False),
            FieldMapping(('children',), MapListOf(NoOpMapper), ('children',), False, True, False),
            FieldMapping(('title',), 'default_action', ('title',), False, False, False),
        ], FromToMapping._mapping_rules)

    def test_map(self):
        from_resource = FromResource(
            # Auto matched
            title="Foo",
            count="42",
            # Excluded
            excluded1=123.4,
            # Mappings
            from_field1="abc",
            from_field2="62",
            from_field3=44,
            from_field4=25,
            same_but_different="def",
            # Custom mappings
            from_field_c1="foo",
            from_field_c2="bar",
            from_field_c3="eek",
            from_field_c4="first-second-third",
            not_auto_c5="do something",
            comma_separated_string="foo,bar,eek",
            child=ChildResource(name='foo'),
            children=[ChildResource(name='foo'), ChildResource(name='bar')]
        )

        to_resource = from_resource.convert_to(ToResource)

        assert 'Foo' == to_resource.title
        assert '42' == to_resource.count
        assert None == to_resource.excluded1
        assert 'abc' == to_resource.to_field1
        assert 62 == to_resource.to_field2
        assert 69 == to_resource.to_field3
        assert 'abc' == to_resource.same_but_different
        assert 'foo-bar-eek' == to_resource.to_field_c1
        assert 'first' == to_resource.to_field_c2
        assert 'second-third' == to_resource.to_field_c3
        assert "DO SOMETHING" == to_resource.not_auto_c5
        assert ['foo', 'bar', 'eek'] == to_resource.array_string

    def test_missing_from_resource(self):
        with pytest.raises(MappingSetupError):
            class _(odin.Mapping):
                to_obj = FakeToResource

    def test_missing_to_resource(self):
        with pytest.raises(MappingSetupError):
            class _(odin.Mapping):
                from_obj = SimpleFromResource

    def test_unknown_from_field_mappings(self):
        with pytest.raises(MappingSetupError):
            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                mappings = (
                    ('unknown_field', None, 'title'),
                )

    def test_unknown_from_field_mappings_multiple(self):
        with pytest.raises(MappingSetupError):
            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                mappings = (
                    (('title', 'unknown_field'), None, 'title'),
                )

    def test_unknown_from_field_custom(self):
        with pytest.raises(MappingSetupError):
            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                @odin.map_field(from_field='unknown_field', to_field='title')
                def multi_to_one(self, *fields):
                    pass

    def test_bad_action_not_callable(self):
        with pytest.raises(MappingSetupError):
            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                mappings = (
                    ('title', 123, 'title'),
                )

    def test_bad_action_not_defined(self):
        with pytest.raises(MappingSetupError):
            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                mappings = (
                    ('title', 'do_transform', 'title'),
                )

    def test_bad_action_defined_not_callable(self):
        with pytest.raises(MappingSetupError):
            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                do_transform = 123

                mappings = (
                    ('title', 'do_transform', 'title'),
                )

    def test_unknown_to_field_mappings(self):
        with pytest.raises(MappingSetupError):
            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                mappings = (
                    ('title', None, 'unknown_field'),
                )

    def test_unknown_to_field_mappings_multiple(self):
        with pytest.raises(MappingSetupError):
            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                mappings = (
                    ('title', None, ('title', 'unknown_field')),
                )

    def test_unknown_to_field_custom(self):
        with pytest.raises(MappingSetupError):
            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                @odin.map_field(from_field='title', to_field='unknown_field')
                def multi_to_one(self, *fields):
                    pass

    def test_bad_mapping(self):
        with pytest.raises(MappingSetupError):
            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                mappings = (
                    'i_forgot', None, 'tuples'
                )

    def test_invalid_list_to_multiple_mapping(self):
        with pytest.raises(MappingSetupError) as cm:
            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                mappings = (
                    ('title', None, ('title', 'name'), True, False, False),
                )
        assert "specifies a to_list mapping, these can only be applied to a" in str(cm.value)

    def test_assignment_with_no_action(self):
        with pytest.raises(MappingSetupError) as cm:
            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                mappings = (
                    (None, None, ('title', 'name'), False, False, False),
                )
        assert "No action supplied for " in str(cm.value)

    def test_from_field_no_field_resolver(self):
        with pytest.raises(MappingSetupError) as cm:
            class _(odin.Mapping):
                from_obj = object
                to_obj = FakeToResource
        assert r"`from_obj`" in str(cm.value)
        assert r"does not have an attribute resolver defined." in str(cm.value)

    def test_to_field_no_field_resolver(self):
        with pytest.raises(MappingSetupError) as cm:
            class _(odin.Mapping):
                from_obj = FakeToResource
                to_obj = object

        assert r"`to_obj`" in str(cm.value)
        assert r"does not have an attribute resolver defined." in str(cm.value)


class TestExecuteMapping(object):
    def test_not_valid_from_resource(self):
        pytest.raises(TypeError, FromToMapping, SimpleFromResource())

    def test_invalid_from_value_count(self):
        with pytest.raises(MappingExecutionError):
            target = FromToMapping(FromResource())
            target._apply_rule((('from_field_c1', 'from_field_c2'), 'one_to_multi', ('title'), False, False, False))

    def test_invalid_to_value_count(self):
        with pytest.raises(MappingExecutionError):
            target = FromToMapping(FromResource(title='Test'))
            target._apply_rule((('title',), 'multi_to_one', ('from_field_c1', 'from_field_c2'), False, False, False))

    def test_apply_single_resource(self):
        f = SimpleFromResource(title="ABC")
        t = SimpleFromTo.apply(f)

        assert isinstance(t, SimpleToResource)
        assert "ABC" == t.title

    def test_apply_single_resource_with_context(self):
        f = SimpleFromResource(title="ABC")
        t = SimpleFromTo.apply(f)

        assert isinstance(t, SimpleToResource)
        assert "ABC" == t.title_count

    def test_apply_multiple_resources(self):
        from_resources = [
            SimpleFromResource(title="Foo"),
            SimpleFromResource(title="Bar"),
            SimpleFromResource(title="Eek"),
        ]

        to_resource_iter = SimpleFromTo.apply(from_resources)
        assert isinstance(to_resource_iter, MappingResult)
        assert ['Foo', 'Bar', 'Eek'] == [t.title for t in to_resource_iter]

    def test_apply_multiple_resources_with_context(self):
        from_resources = [
            SimpleFromResource(title="Foo"),
            SimpleFromResource(title="Bar"),
            SimpleFromResource(title="Eek"),
        ]

        to_resource_iter = SimpleFromTo.apply(from_resources)
        assert isinstance(to_resource_iter, MappingResult)
        assert ['0: Foo', '1: Bar', '2: Eek'] == [t.title_count for t in to_resource_iter]


class TestResourceMappingHelpers(object):
    def test_update_existing(self):
        resource = OldBook(
            name="Bar",  # odin.StringField()
            num_pages=42,  # odin.IntegerField()
            price=12.50,  # odin.FloatField()
            genre="fantasy",  # odin.StringField()
            published=datetime.datetime(2015, 2, 26, 22, 7),  # odin.DateTimeField()
            author=Author(name='123'),  # odin.ObjectAs()
            publisher=Publisher(name="Eek"),  # odin.ObjectAs()
        )
        existing_resource = Book(
            title="Foo",  # odin.StringField()
            num_pages=12,  # odin.IntegerField()
            rrp=12.50,  # odin.FloatField()
            fiction=False,  # odin.BooleanField()
            genre="others",  # odin.StringField()
            published=[datetime.datetime.now()],  # odin.TypedArrayField()
        )

        resource.update_existing(existing_resource)

        assert "Bar" == existing_resource.title
        assert 42 == existing_resource.num_pages
        assert 12.50 == existing_resource.rrp
        assert False == existing_resource.fiction
        assert "fantasy" == existing_resource.genre

    def test_update_filtered_existing(self):
        resource = OldBook(
            name="Bar",  # odin.StringField()
            num_pages=42,  # odin.IntegerField()
            price=12.50,  # odin.FloatField()
            genre="fantasy",  # odin.StringField()
            published=datetime.datetime(2015, 2, 26, 22, 7),  # odin.DateTimeField()
            author=Author(name='123'),  # odin.ObjectAs()
            publisher=Publisher(name="Eek"),  # odin.ObjectAs()
        )
        existing_resource = Book(
            title="Foo",  # odin.StringField()
            num_pages=12,  # odin.IntegerField()
            rrp=12.50,  # odin.FloatField()
            fiction=False,  # odin.BooleanField()
            genre="others",  # odin.StringField()
            published=[datetime.datetime.now()],  # odin.TypedArrayField()
        )

        resource.update_existing(existing_resource, ignore_fields=['num_pages'])

        assert "Bar" == existing_resource.title
        assert 12 == existing_resource.num_pages
        assert 12.50 == existing_resource.rrp
        assert False == existing_resource.fiction
        assert "fantasy" == existing_resource.genre


class ResourceA(odin.Resource):
    class Meta:
        abstract = True

    foo = odin.StringField()


class ResourceB(ResourceA):
    bar = odin.StringField()


class ResourceC(ResourceA):
    eek = odin.StringField()


class ResourceX(odin.Resource):
    class Meta:
        abstract = True

    foo = odin.StringField()


class ResourceY(ResourceX):
    bar = odin.StringField()


class ResourceZ(ResourceX):
    alt = odin.StringField()


class ResourceAToResourceX(odin.Mapping):
    from_obj = ResourceA
    to_obj = ResourceX

    @odin.map_field
    def foo(self, value):
        return "foo: %s" % value


class ResourceBToResourceY(ResourceAToResourceX):
    from_obj = ResourceB
    to_obj = ResourceY


class ResourceCToResourceZ(ResourceAToResourceX):
    from_obj = ResourceC
    to_obj = ResourceZ

    @odin.map_field(from_field='eek')
    def alt(self, value):
        return value


class TestSubClassMapping(object):
    """
    Test the concept of a sub class mapping ie

                Resource A          -->         Mapping A-X         -->           Resource X
                    |                                |                                |
               _____|_____                      _____|_____                      _____|_____
              /           \                    /           \                    /           \
             /            \                   /            \                   /            \
        Resource B    Resource C         Mapping B-Y   Mapping C-Z        Resource Y    Resource Z

    Define a mapping that can handle a list of *Resource A* and *Resource B* objects being mapped by an abstract mapping
    *Mapping A-X* to the corresponding *Resource Y* and *Resource Z* objects.

    """
    def test_abstract_resource_definitions(self):
        assert ['bar', 'foo'] == [f.name for f in ResourceB._meta.fields]
        assert ['eek', 'foo'] == [f.name for f in ResourceC._meta.fields]
        assert ['bar', 'foo'] == [f.name for f in ResourceY._meta.fields]
        assert ['alt', 'foo'] == [f.name for f in ResourceZ._meta.fields]

    def test_abstract_mapping_definitions(self):
        assertMappingEquivalent([
            FieldMapping(('foo',), 'foo', ('foo',), False, False, False),
            FieldMapping(('bar',), 'default_action', ('bar',), False, False, False),
        ], ResourceBToResourceY._mapping_rules)

        assertMappingEquivalent([
            FieldMapping(('eek',), 'alt', ('alt',), False, False, False),
            FieldMapping(('foo',), 'foo', ('foo',), False, False, False),
        ], ResourceCToResourceZ._mapping_rules)

    def test_abstract_mapping(self):
        source = [ResourceB(foo="1", bar="2"), ResourceC(foo="3", eek="4"), ResourceB(foo="5", bar="6")]
        result = ResourceAToResourceX.apply(source)

        assert isinstance(result[0], ResourceY)
        assert isinstance(result[1], ResourceZ)
        assert isinstance(result[2], ResourceY)

    def test_mapping_to_list(self):
        source = [ResourceB(foo="1", bar="2"), ResourceC(foo="3", eek="4"), ResourceB(foo="5", bar="6")]
        result = list(ResourceAToResourceX.apply(source))

        assert isinstance(result[0], ResourceY)
        assert isinstance(result[1], ResourceZ)
        assert isinstance(result[2], ResourceY)

    def test_subs(self):
        assert {
            ResourceB: ResourceBToResourceY,
            ResourceC: ResourceCToResourceZ,
        } == ResourceAToResourceX._subs

    def test_invalid_abstract_mapping_from_obj(self):
        # All sub_class.from_obj should be an sub_class of base_class.from_obj
        with pytest.raises(MappingSetupError) as cm:
            class _(ResourceAToResourceX):
                from_obj = ResourceY
                to_obj = ResourceZ

        assert '`from_obj` must be a subclass of `parent.from_obj`' == str(cm.value)

    def test_invalid_abstract_mapping_to_obj(self):
        # All sub_class.from_obj should be an sub_class of base_class.to_obj
        with pytest.raises(MappingSetupError) as cm:
            class _(ResourceAToResourceX):
                from_obj = ResourceC
                to_obj = ResourceB

        assert '`to_obj` must be a subclass of `parent.to_obj`' in str(cm.value)

    def test_invalid_inherited_type(self):
        with pytest.raises(TypeError) as cm:
            class ResourceD(ResourceA):
                pass
            ResourceAToResourceX.apply(ResourceD())

        assert "`source_resource` parameter must be an instance of" in str(cm.value)




# -*- coding: utf-8 -*-
import pytest

from odin.exceptions import ValidationError

try:
    from odin.contrib.pint.fields import PintField, FloatField
    from odin.contrib.pint.units import registry
except ImportError:
    pass  # Pint is not available to pypy
else:

    class TestPintFields(object):
        # PintField ###############################################################

        def test_pintfield_init(self):
            pytest.raises(ValueError, PintField, None)

        # FloatField ##############################################################

        def test_floatfield_1(self):
            f = FloatField('kWh')
            pytest.raises(ValidationError, f.clean, None)
            pytest.raises(ValidationError, f.clean, 10.2 * registry.meter)
            pytest.raises(ValidationError, f.clean, 'abc')
            assert 10.2 * registry.kilowatt_hour == f.clean(10.2 * registry.kilowatt_hour)
            assert 10.2 * registry['kWh'] == f.clean(10.2)
            assert 10.2 * registry['kWh'] == f.clean((10.2, 'kWh'))
            assert 10.2 * registry['kWh'] == f.clean((10.2, registry.kilowatt_hour))
            assert 10.2 * registry.watt_hour == f.clean(10.2 * registry.watt_hour)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from odin.utils import force_tuple, getmeta

from .base import BaseField

__all__ = ('ConstantField', 'CalculatedField', 'calculated_field', 'MultiPartField')


class VirtualField(BaseField):
    """
    Base class for virtual fields. A virtual fields is treated like any other field during encoding/decoding (provided
    it can be written to).
    """
    data_type_name = None

    def __init__(self, verbose_name=None, verbose_name_plural=None, name=None, data_type_name=None, doc_text='',
                 is_attribute=False, key=False):
        """
        Initialisation of virtual field

        :param verbose_name: Display name of field.
        :param verbose_name_plural: Plural display name of field.
        :param name: Name of the serialised field.
        :param data_type_name: A name for the data type this field returns (used for generating documentation)
        :param doc_text: Documentation for the field, replaces help text
        :param is_attribute: Special flag for codecs that support attributes on nodes (ie XML)
        """
        super(VirtualField, self).__init__(verbose_name, verbose_name_plural, name, doc_text)

        self.data_type_name = data_type_name
        self.is_attribute = is_attribute
        self.key = key

        self.resource = None

    def __get__(self, instance, owner):
        raise NotImplementedError

    def __set__(self, instance, value):
        raise AttributeError("Read only")

    def contribute_to_class(self, cls, name):
        self.set_attributes_from_name(name)
        self.resource = cls
        getmeta(cls).add_virtual_field(self)
        setattr(cls, name, self)


class ConstantField(VirtualField):
    """
    A field that provides a constant value.
    """
    def __init__(self, value, *args, **kwargs):
        super(ConstantField, self).__init__(*args, **kwargs)
        self.value = value

    def __get__(self, instance, owner):
        return self.value


class CalculatedField(VirtualField):
    """
    A field whose that is calculated by an expression.

    The expression should accept a single "self" parameter that is a Resource instance.
    """
    def __init__(self, expr, *args, **kwargs):
        assert callable(expr)
        super(CalculatedField, self).__init__(*args, **kwargs)
        self.expr = expr

    def __get__(self, instance, owner):
        return self.expr(instance)


def calculated_field(method=None, **kwargs):
    """
    Converts an instance method into a calculated field.
    """
    def inner(expr):
        return CalculatedField(expr, **kwargs)
    return inner if method is None else inner(method)


class MultiPartField(VirtualField):
    """
    A field whose value is the combination of several other fields.

    This field should be included after the field that make up the multipart value.
    """
    def __init__(self, field_names, separator='', **kwargs):
        """
        :param field_names: Name(s) of fields to make up key
        :type field_names: str | tuple[str] | list[str]
        :param separator: Separator to use between values.
        :type separator: str
        :param kwargs: Additional kwargs for VirtualField

        """
        kwargs.setdefault('data_type_name', 'String')
        super(MultiPartField, self).__init__(**kwargs)
        self.field_names = force_tuple(field_names)
        self.separator = separator
        self._fields = None

    def __get__(self, instance, owner):
        return self.generate_value(instance)

    def generate_value(self, instance):
        """
        Generate a key based on other values.
        """
        values = [f.prepare(f.value_from_object(instance)) for f in self._fields]
        return self.separator.join(str(v) for v in values)

    def on_resource_ready(self):
        # Extract reference to fields
        meta = getmeta(self.resource)
        try:
            self._fields = tuple(meta.field_map[name] for name in self.field_names)
        except KeyError as ex:
            raise AttributeError("Attribute {0} not found on {1!r}".format(ex, self.resource))



# -*- coding: utf-8 -*-
import pytest
import odin


class MultiPartResource(odin.Resource):
    id = odin.IntegerField()
    code = odin.StringField()
    two_parts = odin.MultiPartField(('id', 'code'), separator=':')


class TestFields(object):
    def test_multipartfield__get_value(self):
        target = MultiPartResource(id=42, code='29A')

        assert '42:29A' == target.two_parts

    def test_multipartfield__unknown_fields(self):
        with pytest.raises(AttributeError) as result:
            class BadMultiPartResource(odin.Resource):
                id = odin.IntegerField()
                two_parts = odin.MultiPartField(('id', 'code'), separator=':')

        assert str(result.value).startswith("Attribute 'code' not found")
